@@ -110,12 +110,38 @@ hasSelection (LPDISPATCH explorer)
     }
 
   int count = get_oom_int (selection, "Count");
-  gpgol_release (selection);
 
   if (count)
     {
-      return true;
+      LPDISPATCH selectitem = NULL, mailitem = NULL;
+      bool selected = hasMailitemEventReadBeenCalled ();
+      if (count == 1)
+        {
+          // If we call this on a selection with more items
+          // Outlook sends an ItemLoad event for each mail
+          // in that selection.
+          selectitem = get_oom_object (selection, "Item(1)");
+          mailitem = get_object_by_id (selectitem, IID_MailItem);
+          if (!mailitem)
+            {
+              log_debug ("%s:%s: New selection is no mail.",
+                SRCNAME, __func__);
+              selected = false;
+            }
+
+          gpgol_release (mailitem);
+          gpgol_release (selectitem);
+        }
+      else
+        {
+          selected = false; // We can't show the security level for more than one => show insecure
+        }
+
+      gpgol_release (selection);
+      return selected;
     }
+
+  gpgol_release (selection);
   return false;
 }
 
@@ -160,6 +186,14 @@ start_watchdog (LPVOID arg)
 static void
 changeSeen (LPDISPATCH explorer)
 {
+  auto view = get_oom_object_s (explorer, "CurrentView");
+  if (view && get_object_name_s (view.get ()) == "_PeopleView")
+    {
+      log_oom ("Selection change in people view. Invalidating.");
+      gpgoladdin_invalidate_ui ();
+      return;
+    }
+
   gpgol_lock (&explorer_map_lock);
 
   auto it = s_explorerMap.find (explorer);
